#include "SDUpdater.h"

SDUPDATER::SDUPDATER() {
	running_partition = esp_ota_get_running_partition();
	update_partition  = esp_ota_get_next_update_partition(NULL);
}

SDUPDATER::~SDUPDATER() {}

void SDUPDATER::rollback()
{
	esp_ota_set_boot_partition (update_partition);
}

void SDUPDATER::begin(WiFiServer *tnserver, const char *firmware_path)
{
	tnServer  = tnserver;
	firm_path = firmware_path;
	pt        = stack;
	*pt       = 0x00;
	filenum   = 0;
}

boolean SDUPDATER::available()
{
	unsigned char negotiation[25] = {
		0xFF,0xFB,0x03, // IAC WILL suppress_go_ahead
		0xFF,0xFD,0x01, // IAC DO   echo
		0xFF,0xFD,0x1F, // IAC DO   window_size
		0xFF,0xFB,0x05, // IAC WILL status
		0xFF,0xFD,0x21, // IAC DO   remote_flow_control
		0xFF,0xFD,0x18, // IAC DO   terminal_type
		0xFF,0xFD,0x03, // IAC DO   suppress_go_ahead
		0xFF,0xFB,0x01, // IAC WILL echo
	};

	if(!client) {
		client = tnServer->available();
		if(client) {
			if (client.connected()) {
				while (client.available()) getClientChar();
				for(int i=0;i<24;++i) client.write(negotiation[i]);
				while (client.available()) getClientChar();
			}
		}
	} else if (!client.connected()) {
		client.stop();
		client = tnServer->available();
		if(client) {
			if (client.connected()) {
				while (client.available()) getClientChar();
				for(int i=0;i<24;++i) client.write(negotiation[i]);
				while (client.available()) getClientChar();
			}
		}
	}

	if(client) {
		if (client.connected()) {
			while (client.available()) {
				ch = getClientChar();
				if (ch==0xFF) {
					ch = getClientChar();
					switch(ch) {
						case 0xF0: 
							break;
						case 0xFA: 
							for(int i=0;i<5;++i) {
								ch = getClientChar();
							}
							break;
						default:
							ch = getClientChar();
							break;
					}
				} else {
					return true;
				}
			}
			ch = NULL;
			return true;
		}
	}
	return false;
}

int SDUPDATER::read()
{
	return ch;
}

void SDUPDATER::print(const char* format, ...)
{
	va_list args;
	va_start(args, format);
	vsprintf(buf,format,args);
	va_end(args);

	if (tnServer) {
		if (available()) {
			for (int len = 0; buf[len]; len++) client.write(buf[len]);
			return;
		}
	}
}

void SDUPDATER::write(char ch)
{
	if (tnServer) {
		client.write(ch);
		return;
	}
}

int SDUPDATER::getClientChar()
{
	while(1) if(client.available()) return client.read();
}

uint32_t SDUPDATER::running_address()
{
	return running_partition->address;
}

uint32_t SDUPDATER::update_address()
{
	return update_partition->address;
}

void SDUPDATER::update()
{
	int  selno = 0;
	char ch;
	size_t updateSize;

	print("Firmware List\n");

	bin = SD.open(firm_path);
	getFirmwareName(bin, 0);
	bin.close();

	if (filenum == 0) return;

	while(1) {
		print("Select File No. => ");
		while(1) {
			if (available()) {
				ch = read();
				if ((ch>=0x30)&&(ch<=0x39)) {
					if (selno == 0) {
						selno = ch - 0x30;
					} else {
						selno = selno * 10 + (ch - 0x30);
					}
					write(ch);
				} else {
					if ((ch==0x0d)||(ch==0x0a)) {
						write(ch);
						break;
					}
				}
			}
		}
		if ((selno>0)&&(selno<=filenum)) {
			break;
		} else {
			selno = 0;
		}
	}
	print("\n");

	pt = stack;
	while(--selno) {
		while(*pt) pt++;
		pt++;
	}
	sprintf(firmwareName,"%s/%s.bin",firm_path,pt);

	File updateBin = SD.open(firmwareName);
	if (updateBin) {
		updateSize = updateBin.size();
		print("%s %d bytes\n",firmwareName, updateSize);
		print("update partition address:0x%08x(0x%08x)\n", update_partition->address, update_partition->size);

		if (Update.begin(updateSize)) {      
			size_t written = Update.writeStream(updateBin);
			if (written == updateSize) {
				print("Written : %d byte successfully\n",written);
			} else {
				print("Written only : %d / %d. Retry?\n",written, updateSize);
			}
			if (Update.end()) {
				if (Update.isFinished()) {
					print("Update successfully completed. Rebooting.\n");
				} else {
					print("Update not finished? Something went wrong!\n");
				}
			} else {
				print("Error Occurred. Error #: %s\n",Update.getError());
			}
		} else {
			Serial.println("Not enough space to begin SDUPDATER");
		}

		updateBin.close();
		delay(3000);
		ESP.restart();

	} else {
		print("failed to open %s\n",firmwareName);
	}
}

void SDUPDATER::restart()
{
	ESP.restart();
}

void SDUPDATER::getFirmwareName(File dir, int numTabs)
{
	dir.rewindDirectory();

	while(true) {
		File entry =  dir.openNextFile();
		if (! entry) break; // no more files
		if (!entry.isDirectory()) {
			strcpy(pt,entry.name());
			for (ckpt=pt;*ckpt!='.';ckpt++);
			if (memcmp(ckpt+1,"bin",3)==0) {
				filenum++;
				*ckpt = 0x00;
				pt = ckpt + 1;
			}
		} else {
			print("/\n");
			getFirmwareName(entry, numTabs+1);
		}
		entry.close();
	}

	pt = stack;
	for (int i=1; i<=filenum; i++) {
		print("%d:%s\n",i,pt);
		while(*(++pt));
		pt++;
	}
}
